#include "testApp.h"

// application to count change in provided images
// trying to use basic pixel count first, then maybe countours

#define kMargin 20

//--------------------------------------------------------------
void testApp::setup(){	 
    threshold = 360;
    
    // a vector of all the images
    ofImage tmpimg;
    for(int i=1;i<9;i++){
        stringstream ipath;
        ipath << "images/image" << i << ".jpg";
        tmpimg.loadImage(ipath.str());
        coinImages.push_back(tmpimg);
    }
    // and the huge image:
    tmpimg.loadImage("images/super_change.jpg");
    coinImages.push_back(tmpimg);

    selectedImage = 0;
    
    // a list of objects to track what we find:
    list<Object> objects;
    
    // previously calibrated coin sizes from known images.
    coinSizes.push_back(5610); // penny
    coinSizes.push_back(5689); // nickel
    coinSizes.push_back(4346); // dime
    coinSizes.push_back(7330); // quarter
    
    initTextures();
    doDetection(coinImages[selectedImage]);
    totalValue = countChange();
    
    ttf_big.loadFont("type/frabk.ttf",20);
    ttf_sm.loadFont("type/frabk.ttf", 14);
    ttf_tiny.loadFont("type/frabk.ttf", 9);
    

}

//--------------------------------------------------------------
void testApp::update(){
    ofBackground(255,255,255);	
}

//--------------------------------------------------------------
void testApp::draw(){
    
    glPushMatrix(); 
    glTranslatef(kMargin,kMargin+60,0); 
    
    // display the image
    ofSetColor(255,255,255);
//    if(displayThreshold)
//        thresholdTexture.draw(0,0);
//    else
        displayImage.draw(0,0);
    
    // now compute/display midpoints   
    if(displayTargets){
        list<Object>::iterator it;
        for ( it=objects.begin() ; it != objects.end(); it++ ){
            Object ob = *it;
            
            // draw stuff to represent the midpoints.
            int x = ob.xsum/ob.count;
            int y = ob.ysum/ob.count;
            
            // adjust for image scale:
            x = (int) x*displayRatio;
            y = (int) y*displayRatio;
            
            ofSetColor(0xff0000);
            makeTarget(x,y);    

        }
    }
    glPopMatrix();
    
    // display the title / instructions
    int yMargin = kMargin+5;
    ofSetRectMode(OF_RECTMODE_CORNER);
    ofSetColor(50, 50, 50);
    ttf_big.drawString("Identifying and counting change", kMargin, yMargin);
    ttf_tiny.drawString("[Press 1-9 to change image]", kMargin, yMargin+17);
    
    // display the counted value
    stringstream msg;
    msg << "Value of change found: ";
    msg << totalValue;
    if(totalValue > 1)
        msg << " cents";
    else
        msg << " cent";
    
    ttf_sm.drawString(msg.str(), kMargin, yMargin+40);

}


void testApp::initTextures(){
    imgWidth = coinImages[selectedImage].width;
    imgHeight = coinImages[selectedImage].height;
    
    thresholdTexture.allocate(imgWidth, imgHeight, GL_LUMINANCE);
    thresholdPixels = new unsigned char[imgWidth*imgHeight];
    
    objectTexture.allocate(imgWidth, imgHeight, GL_LUMINANCE);
    objectPixels = new unsigned char[imgWidth * imgHeight];
    
    //load in the current image, scaling if necessary for display
    displayImage = coinImages[selectedImage];
    if(displayImage.width > ofGetWidth()-kMargin*2){
        float aspect = displayImage.height/(float)displayImage.width;
        int newWidth = ofGetWidth()-kMargin*2;
        displayImage.resize(newWidth, newWidth * aspect);
    }
    displayRatio = displayImage.width / (float)coinImages[selectedImage].width;
}

//--------------------------------------------------------------
// basic seed fill function - would be faster to work with pixel array directly
void testApp::seedFill(int x, int y, int label){
    int pixelLoc = x+(y*imgWidth);
    
    //background, or already scanned return:
    if(objectPixels[pixelLoc]==0||objectPixels[pixelLoc]==label) return; 
    
    if(objectPixels[pixelLoc]==255){
        objectPixels[pixelLoc] = label;// setPixel(x,y, label);
        seedFill(x,y+1,label); //north
        seedFill(x,y-1,label); //south
        seedFill(x+1,y,label); //west
        seedFill(x-1,y,label); //east
    }
}    

void testApp::doDetection(ofImage img){
    // do the thresholding.
    // scan through each pixel and test the value.
    int w = img.width;
    int h = img.height;
    unsigned char * pixels = img.getPixels();
    
    int totalPixels = imgWidth*imgHeight;
    for(int i=0;i<totalPixels;i++){
        int averageVal;
        averageVal += pixels[i*3];
        averageVal += pixels[(i*3)+1];
        averageVal += pixels[(i*3)+2];
        averageVal /= 3;
        
        if(averageVal < threshold)
            thresholdPixels[i] = 255;
        else
            thresholdPixels[i] = 0;   
        
        objectPixels[i] = thresholdPixels[i];
    }
    
    thresholdTexture.loadData(thresholdPixels, imgWidth, imgHeight, GL_LUMINANCE);
    
    // seed fill object detection:
    // set the initial label color:
    // valid values are 1-254. 0 = background, 255 = unlabeled object
    int label = 1;
    
    // remove the old objects in the list:
    objects.clear();
    
    for(int i=1;i<imgWidth-1;i++){
        for(int j=1;j<imgHeight-1;j++){
            // limited to 253 labeled objects
            if(objectPixels[i+(j*imgWidth)] == 255 && label < 254) {
                // look for unlabeled object pixels
                Object ob = {
                    label,0,0,0
                };
                    objects.push_back(ob);
                
                seedFill(i,j,label++);
            }
        }
    }
    
    objectTexture.loadData(objectPixels, imgWidth, imgHeight, GL_LUMINANCE);
    
    
    // Feature Recognition. Look for object midpoints.
    for(int i=1;i<imgWidth-1;i++){
        for(int j=1;j<imgHeight-1;j++){
            int thisPixel = objectPixels[i+(j*imgWidth)];
            if(thisPixel > 0){
                // this is a tagged object
                // look for it's label and add to the list
                list<Object>::iterator it;
                for ( it=objects.begin() ; it != objects.end(); it++ ){
                    Object ob = *it;
                    if(ob.label == thisPixel){
                        ob.xsum += i;
                        ob.ysum += j;
                        ob.count = ob.count+1;
                        
                        objects.insert(it,ob);
                        objects.erase(it);
                    }
                }
            } 
        }
    }
    

    // filter out small objects (noise)
    list<Object>::iterator it;
    for ( it=objects.begin() ; it != objects.end(); it++ ){
        Object ob = *it;
        if(ob.count < 20){
            objects.erase(it);
            if(it == objects.end())
                break;
        }
    }
    
    for ( it=objects.begin() ; it != objects.end(); it++ ){
        Object ob = *it;
        int value = ob.count;
        
        for(int i=0;i<4;i++){
            int diff = abs(coinSizes[i] - ob.count);
            if(diff<50){
                // found the currency
                switch(i){
                    case 0:
                        ob.value = 1;
                        break;
                    case 1:
                        ob.value = 5;
                        break;
                    case 2:
                        ob.value = 10;
                        break;
                    case 3:
                        ob.value = 25;
                        break;
                }
                
                objects.insert(it,ob);
                objects.erase(it);
                
                break;
            }
        }
    }
}

void testApp::makeTarget(int _x, int _y){
    // draw a cross hair
    int dia = 30;
    ofEnableAlphaBlending();
    ofNoFill();
    
    // (dark) cross hairs, with break in the middle
    ofSetColor(20, 20, 20, 100);
    for(int i=0;i<4;i+=2){
        ofLine(_x, _y-dia/2+(i*dia/3), _x, _y-dia/2+((i+1)*dia/3));
        ofLine(_x-dia/2+(i*dia/3),_y, _x-dia/2+((i+1)*dia/3),_y);
    }
    ofRect(_x-dia/3, _y-dia/3, 20,20);
    
    // (light) cross hairs, with break in the middle
    _x++;_y++;
    ofSetColor(200, 200, 200, 100);
    for(int i=0;i<4;i+=2){
        ofLine(_x, _y-dia/2+(i*dia/3), _x, _y-dia/2+((i+1)*dia/3));
        ofLine(_x-dia/2+(i*dia/3),_y, _x-dia/2+((i+1)*dia/3),_y);
    }
    ofRect(_x-dia/3, _y-dia/3, 20,20);
    
    ofDisableAlphaBlending();
    ofFill();
}


int testApp::countChange(){
    // calculate the value of the found objects
    int values = 0;
    list<Object>::iterator it;
    for ( it=objects.begin() ; it != objects.end(); it++ ){
        Object ob = *it;
        values += ob.value;
    }
    return values;
}


//--------------------------------------------------------------
void testApp::keyPressed  (int key){
    list<Object>::iterator it;
    
	switch(key){
        case 'p':
            // print out the sizes of the found objects
            for ( it=objects.begin() ; it != objects.end(); it++ ){
                Object ob = *it;
                cout << "object: " << ob.label << ", size: " << ob.count << ", value: " << ob.value << endl;
            }
            break;
            
        case 'c':
            // calculate the tally and print to the console
            totalValue =  countChange();
            cout << "tally: " << totalValue << " cents" << endl;
            break;
            
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            selectedImage = key-48-1; //ascii to decimal, then -1 for zero-based vector
            initTextures();
            doDetection(coinImages[selectedImage]);
            totalValue = countChange();
            break;
            
        case 't':
            // toggle thresholding
            displayThreshold = !displayThreshold;
            break;
            
        case 'T':
            displayTargets = !displayTargets;
            break;
    }
}

//--------------------------------------------------------------
void testApp::keyReleased  (int key){ 
	
}

//--------------------------------------------------------------
void testApp::mouseMoved(int x, int y ){
}

//--------------------------------------------------------------
void testApp::mouseDragged(int x, int y, int button){
//    // update the threshold value
//    threshold = (int)255*(y/(float)ofGetHeight());
//    cout << threshold << endl; // 360?
}

//--------------------------------------------------------------
void testApp::mousePressed(int x, int y, int button){
}

//--------------------------------------------------------------
void testApp::mouseReleased(){

}
